unit Color;

interface
uses Graphics;

type
  TRGB = record
    B, G, R: Byte;
  end;

  TRGBArray = array[0..400000] of TRGB;
  PRGBArray = ^TRGBArray;

  TYUV = record
    Y, U, V: Byte;
  end;

  TYUVArray = array[0..400000] of TYUV;
  PYUVArray = ^TYUVArray;

  TByteTriple = array[0..2] of Byte;
  TBTArray = array[0..400000] of TByteTriple;
  PBTArray = ^TBTArray;

  function RGB2YUV(const RGB: TRGB): TYUV;
  function YUV2RGB(const YUV: TYUV): TRGB;
  procedure MonochromeFilter(const Src: TBitmap; U, V: Integer);

implementation

procedure MonochromeFilter(const Src: TBitmap; U, V: Integer);
var
  X, Y, W: Integer;
  P: PRGBArray;
  Table: array[0..255] of TRGB;
  YUV: TYUV;
begin
  YUV.U := U;
  YUV.V := V;

  for Y := 0 to $FF do
  begin
    YUV.Y := Y;
    Table[Y] := YUV2RGB(YUV);
  end;

  W := Src.Width - 1;
  for Y := 0 to Src.Height - 1 do
  begin
    P := Src.ScanLine[Y];
    for X := 0 to W do
    begin
      with P[X] do
      begin
        P[X] := Table[(627049 * R + 1231028 * G + 239075 * B + 1048576) shr 21];
      end;
    end;
  end;
end;

function RGB2YUV(const RGB: TRGB): TYUV;
begin
  with RGB, Result do
  begin
    Y := (627049 * R + 1231028 * G + 239075 * B + 1048576) shr 21;
    U := (1183494 * (B - Y) + 268435456) shr 21;
    V := (1495829 * (R - Y) + 268435456) shr 21;
  end;
end;

function YUV2RGB(const YUV: TYUV): TRGB;
var
  R, G, B: Integer;
begin
  with YUV do
  begin
    R := (Y shl 21 + 2940207 * V - 373827830) div 2097152;
    G := (Y shl 21 - 721706 * U - 1497652 * V + 284016801) div 2097152;
    B := (Y shl 21 + 3716153 * U - 472760975) div 2097152;
  end;

  if R < 0 then R := 0 else if R > $FF then R := $FF;
  if G < 0 then G := 0 else if G > $FF then G := $FF;
  if B < 0 then B := 0 else if B > $FF then B := $FF;

  Result.R := R;
  Result.G := G;
  Result.B := B;
end;

end.
