unit main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Menus,
  ExtCtrls, StdCtrls, IniFiles, ShellApi, Color, MyUtils, ComCtrls, stretchf,
  Dialogs;

type
  TMainForm = class(TForm)
    UVImage: TImage;
    MainMenu: TMainMenu;
    HelpMenu: TMenuItem;
    YImage: TImage;
    FileMenu: TMenuItem;
    ExitMenu: TMenuItem;
    UEdit: TEdit;
    VEdit: TEdit;
    ULabel: TLabel;
    VLabel: TLabel;
    AboutMenu: TMenuItem;
    N1: TMenuItem;
    ConfigMenu: TMenuItem;
    ResetMenu: TMenuItem;
    WebMenu: TMenuItem;
    MailMenu: TMenuItem;
    StatusBar: TStatusBar;
    ExecuteButton: TButton;
    OpenMenu: TMenuItem;
    OpenDialog: TOpenDialog;
    N2: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure UEditChange(Sender: TObject);
    procedure VEditChange(Sender: TObject);
    procedure AboutMenuClick(Sender: TObject);
    procedure UVImageMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure UVImageMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure UVImageMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure ResetMenuClick(Sender: TObject);
    procedure WebMenuClick(Sender: TObject);
    procedure MailMenuClick(Sender: TObject);
    procedure ExecuteButtonClick(Sender: TObject);
    procedure ExitMenuClick(Sender: TObject);
    procedure OpenMenuClick(Sender: TObject);
  private
    FDownNow: Boolean;
    Src: TBitmap;
    SrcName: TFileName;
    procedure DrawUVImage;
    procedure DrawYImage;
    procedure WMDropFiles(var Msg: TWMDropFiles); message WM_DROPFILES;
    procedure AppHint(Sender: TObject);
    procedure OpenImage(const FileName: string);
  public
    property DownNow: Boolean read FDownNow write FDownNow;
  end;

  TConfig = class
  private
    FU: Byte;
    FV: Byte;
    function UDefault: Byte;
    function VDefault: Byte;
  public
    constructor Create; virtual;
    destructor Destroy; override;
  published
    property U: Byte read FU write FU;
    property V: Byte read FV write FV;
  end;

var
  MainForm: TMainForm;
  Config: TConfig;

implementation

{$R *.DFM}

{ TConfig }

constructor TConfig.Create;
var
  IniFile: TIniFile;
begin
  IniFile := TIniFile.Create(LowerCase(ChangeFileExt(Application.ExeName, '.ini')));
  try
    //for All Applications
    with MainForm, IniFile do
    begin
      Left    := ReadInteger('Window', 'Left', Left);
      Top     := ReadInteger('Window', 'Top',  Top);
      //Width   := ReadInteger('Window', 'Width', Width);
      //Height  := ReadInteger('Window', 'Height',  Height);
    end;

    //for This Application
    with IniFile do
    begin
      U := ReadInteger('Color', 'U', UDefault);
      V := ReadInteger('Color', 'V', VDefault);
    end;

  finally
    IniFile.Free;
  end;
end;

destructor TConfig.Destroy;
var
  IniFile: TIniFile;
begin
  IniFile := TIniFile.Create(LowerCase(ChangeFileExt(Application.ExeName, '.ini')));
  try
    //for All Applications
    with MainForm, IniFile do
    begin
      WriteInteger('Window', 'Left', Left);
      WriteInteger('Window', 'Top',  Top);
      //WriteInteger('Window', 'Width', Width);
      //WriteInteger('Window', 'Height',  Height);
    end;

    //for This Application
    with IniFile do
    begin
      WriteInteger('Color', 'U', U);
      WriteInteger('Color', 'V', V);
    end;
  finally
    IniFile.Free;
  end;
end;

function TConfig.UDefault: Byte;
begin
  Result := 117;
end;

function TConfig.VDefault: Byte;
begin
  Result := 137;
end;

procedure TMainForm.AppHint(Sender: TObject);
begin
  StatusBar.SimpleText := Application.Hint;
end;


procedure TMainForm.DrawUVImage;
var
  U, V: Integer;
  YUV: TYUV;
  P: PRGBArray;
begin
  YUV.Y := 127;
  for V := 0 to $FF do
  begin
    P := UVImage.Picture.Bitmap.ScanLine[V];
    YUV.V := V;
    for U := 0 to $FF do
    begin
      YUV.U := U;
      P[U] := YUV2RGB(YUV);
    end;
  end;
  UVImage.Invalidate;
end;

procedure TMainForm.DrawYImage;
var
  Y, W, X: Integer;
  YUV: TYUV;
  RGB: TRGB;
  P: PRGBArray;
begin
  YUV.U := Config.U;
  YUV.V := Config.V;
  W := YImage.Picture.Bitmap.Width;
  for Y := 0 to $FF do
  begin
    P := YImage.Picture.Bitmap.ScanLine[Y];
    YUV.Y := $FF - Y;
    RGB := YUV2RGB(YUV);
    for X := 0 to W - 1 do
    begin
      P[X] := RGB;
    end;
  end;
  YImage.Invalidate;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  Config := TConfig.Create;

  DragAcceptFiles(MainForm.Handle, True);
  Application.OnHint := AppHint;

  with UVImage.Picture.Bitmap do
  begin
    Width := UVImage.Width;
    Height := UVImage.Height;
    PixelFormat := pf24bit;
  end;
  DrawUVImage;

  with YImage.Picture.Bitmap do
  begin
    Width := YImage.Width;
    Height := YImage.Height;
    PixelFormat := pf24bit;
  end;

  UEdit.Text := IntToStr(Config.U);
  VEdit.Text := IntToStr(Config.V);

  DownNow := False;
  Src := TBitmap.Create;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  Src.Free;
  Config.Free;
end;

procedure TMainForm.UEditChange(Sender: TObject);
var
  U: Integer;
begin
  U := StrToIntDef(UEdit.Text, -1);
  if (U >= 0) and (U <= $FF) then Config.U := U;
  DrawYImage;
end;

procedure TMainForm.VEditChange(Sender: TObject);
var
  V: Integer;
begin
  V := StrToIntDef(VEdit.Text, -1);
  if (V >= 0) and (V <= $FF) then Config.V := V;
  DrawYImage;
end;

procedure TMainForm.AboutMenuClick(Sender: TObject);
begin
  AboutBox;
end;

procedure TMainForm.UVImageMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  DownNow := True;
  UEdit.Text := IntToStr(X);
  VEdit.Text := IntToStr(Y);
  DrawYImage;
end;

procedure TMainForm.UVImageMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  DownNow := False;
end;

procedure TMainForm.UVImageMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
begin
  if DownNow then
  begin
    if (X >= 0) and (X <= $FF) then UEdit.Text := IntToStr(X);
    if (Y >= 0) and (Y <= $FF) then VEdit.Text := IntToStr(Y);
    DrawYImage;
  end;
end;

procedure TMainForm.ResetMenuClick(Sender: TObject);
begin
  UEdit.Text := IntToStr(Config.UDefault);
  VEdit.Text := IntToStr(Config.VDefault);
  DrawYImage;
end;

procedure TMainForm.WebMenuClick(Sender: TObject);
begin
  GoWeb;
end;

procedure TMainForm.MailMenuClick(Sender: TObject);
begin
  SendMail;
end;

procedure TMainForm.WMDropFiles(var Msg: TWMDropFiles);
var
  FileName: string;
begin
  try
    SetLength(FileName, MAX_PATH + 1);
    DragQueryFile(Msg.Drop, 0, PChar(FileName), MAX_PATH);
    SetLength(FileName, StrLen(PChar(FileName)));
    OpenImage(FileName);
  finally
    DragFinish(Msg.Drop);
  end;
end;

procedure TMainForm.ExecuteButtonClick(Sender: TObject);
var
  Bitmap: TBitmap;
  FileName: TFileName;
begin
  if SrcName = '' then
  begin
    StatusBar.SimpleText := '摜ǂݍ܂Ă܂';
    Exit;
  end;

  Bitmap := TBitmap.Create;
  try
    Bitmap.Assign(Src);
    MonochromeFilter(Bitmap, Config.U, Config.V);
    UVImage.Picture.Bitmap.Assign(Bitmap);
    Stretch(UVImage.Picture.Bitmap, UVImage.Width, UVImage.Height, nil);
    Application.ProcessMessages;
    if Application.MessageBox('ۑ܂H', 'mF', MB_YESNO or MB_ICONQUESTION) = IDYES then
    begin
      FileName := ExtractFilePath(SrcName) + '$' + ExtractFileName(SrcName);
      Bitmap.SaveToFile(FileName);
      StatusBar.SimpleText := 'ۑ ' + FileName;
    end;
  finally
    DrawUVImage;
    Bitmap.Free;
  end;
end;

procedure TMainForm.ExitMenuClick(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.OpenMenuClick(Sender: TObject);
begin
  if OpenDialog.Execute then OpenImage(OpenDialog.FileName);
end;

procedure TMainForm.OpenImage(const FileName: string);
  function LoadImage(const Bitmap: TBitmap; const FileName: TFileName): Boolean;
  begin
    Result := False;
    if LowerCase(ExtractFileExt(FileName)) = '.bmp' then
    begin
      Bitmap.LoadFromFile(FileName);
      if Bitmap.PixelFormat <> pf24bit then
      begin
        Bitmap.PixelFormat := pf24bit;
        Bitmap.ReleasePalette;
      end;
      Result := True;
    end;
  end;
begin
  if LoadImage(Src, FileName) then
  begin
    SrcName := FileName;
    StatusBar.SimpleText := 'ǂݍ݊ ' + FileName;
  end;
end;

end.
